/*
 * Copyright (c) 2013, 
 * Sangoma Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

#include "sample_gsm.h"

/********************************************************************************/
/*                                                                              */
/*                           function declaration                               */
/*                                                                              */
/********************************************************************************/
#ifndef __WINDOWS__
void on_sigint(int signal);
void on_sigusr(int signal);
#endif

int sample_app_main_menu(void);
int sample_app_send_sms(span_data_t * span_data);
int sample_app_call_req(span_data_t * span_data);
int sample_app_print_status(span_data_t * span_data);
span_data_t *sample_app_span_menu(void);

/* Set SDK_DEBUGGING since this is a sample application,
 * This should be turned off in Production */
#define SDK_DEBUGGING 1

/********************************************************************************/
/*                                                                              */
/*                              global parameters                               */
/*                                                                              */
/********************************************************************************/
int g_running = 1;
int g_num_threads = 0;

span_data_t SPANS[FTDM_MAX_SPANS_INTERFACE + 1] = { {0} };
sangoma_freetdm_init_t sample_app_ftdm_init;

/********************************************************************************/
/*                                                                              */
/*                              implementation                                  */
/*                                                                              */
/********************************************************************************/

#ifndef __WINDOWS__
void on_sigusr(int signal)
{
	switch (signal) {
	case SIGUSR1:
		fprintf(stdout, "User interrupted... \n");
		break;

	case SIGUSR2:
		fprintf(stderr, "User interrupted... \n");
		break;
	}

	fprintf(stdout, "SANGOMA GSM MODULE Application is Getting closed\n");
	ftdm_global_destroy();
	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");
	getchar();
	fprintf(stdout, "Press enter... \n");
	exit(EXIT_FAILURE);
}

void on_sigint(int signal)
{
	g_running = 0;

	fprintf(stdout, "SIGINT recieved ...\n");
	fprintf(stdout, "SANGOMA GSM MODULE Application is Getting closed\n");
	ftdm_global_destroy();
	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                      GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "Press Enter key to exit... \n");
	getchar();
	exit(EXIT_FAILURE);
}
#endif

/****************************************
 FreeTDM Event Callbacks
 ****************************************/

int sample_app_hwprobe_callback(void *sangoma_port_obj, int port_id, void *user_obj)
{
	int card_type = sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj);
	int err;

	/* Filter out all other cards just look for W400 */
	if (card_type != AFT_ADPTR_W400) {
		/* Indicate that we are skipping configuration of this card */
		return 1;
	}

	SPANS[port_id].init = 1;
	SPANS[port_id].sangoma_port_obj = sangoma_port_obj;

	err = sangoma_config_port_stop(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
	}

	err = sangoma_config_port_save(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
		return -1;
	}

	err = sangoma_config_port_start(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to start sangoma port config\n");
		return -1;
	}

	/* Create spans in FreeTDM */
	err = sangoma_freetdm_create_span(sangoma_port_obj, &SPANS[port_id].ftdm_span, NULL);
	if (err) {
		fprintf(stderr, "Error: Failed to create freetdm span %i\n", port_id);
		return -1;
	}

	/* Loads the configuration with respect to GSM module */
	if (sangoma_freetdm_span_config_and_start(sangoma_port_obj, SPANS[port_id].ftdm_span, NULL, NULL)) {
		fprintf(stderr, "Error: Failed to load and configure freetdm span %i\n", port_id);
		return -1;
	}

	return 0;
}

int sample_app_send_sms(span_data_t * span_data)
{
	char called[128];
	char text[512];

	while (1) {
		fprintf(stdout, "\nPhone Number: ");
		if (fgets(called, 128, stdin) == NULL) {
			fprintf(stdout, "\nError try again");
		} else {
			break;;
		}
	}

	called[strlen(called) - 1] = '\0';

	while (1) {
		fprintf(stdout, "\nSMS Text: ");
		if (fgets(text, 512, stdin) == NULL) {
			fprintf(stdout, "\nError try again");
		} else {
			break;;
		}
	}

	text[strlen(text) - 1] = '\0';

	return sangoma_freetdm_send_sms(span_data->ftdm_span, called, text);
}

int sample_app_at(span_data_t * span_data)
{
	char command[1024];

	while (1) {
		fprintf(stdout, "\nPlease type in AT command (max 1024 char): ");
		if (fgets(command, 1024, stdin) == NULL) {
			fprintf(stderr, "\nError try again");
		} else {
			break;
		}
	}

	command[strlen(command) - 1] = '\0';
	
	return sangoma_freetdm_send_at(span_data->ftdm_span, command);
}

int sample_app_call_req(span_data_t * span_data)
{
	char called[128];
	char calling[128];

	while (1) {
		fprintf(stdout, "\nCalled Number: ");
		if (fgets(called, 128, stdin) == NULL) {
			fprintf(stdout, "\nError try again");
		} else {
			break;;
		}
	}

	called[strlen(called) - 1] = '\0';

	while (1) {
		fprintf(stdout, "\nCalling Number: ");
		if (fgets(calling, 128, stdin) == NULL) {
			fprintf(stdout, "\nError try again");
		} else {
			break;;
		}
	}

	calling[strlen(calling) - 1] = '\0';

	return sangoma_freetdm_call_req(span_data->ftdm_span, called, calling);
}

int gsm_main_menu(void)
{
	char choice[6];
	span_data_t *span_data;

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "		WELCOME TO SANGOMA GSM MODULE             \n");
	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "Please select any one of the below options :\n");

	span_data = sample_app_span_menu();

	fprintf(stdout, "Select Action\n");
	fprintf(stdout, "1. Send SMS\n2. Call To A Number\n3. Print Span Information\n4. Send AT command\n5. Exit\n");
	fprintf(stdout, "Enter your choice[1-4]: ");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			fprintf(stderr, "Invalid Input Choice must must be [1-4]...\n");
			return 0;
		}

		choice[1] = '\0';

		/* Provides option to user to perform various operations like:
		 * 1) Send SMS
		 * 2) Print Span related information
		 * 3) Exit from application
		 */
		switch (choice[0]) {
		case '1':
			sample_app_send_sms(span_data);
			break;

		case '2':
			sample_app_call_req(span_data);
			break;

		case '3':
			while (sample_app_print_status(span_data)) {
				continue;
			}
			break;
			
		case '4':
			sample_app_at(span_data);
			break;

		case '5':
			fprintf(stdout, "GSM Sample Application is getting closed...\n");
			fprintf(stdout, "\n\n\n=========================================================\n");
			fprintf(stdout, "                      GOOD BYE                            \n");
			fprintf(stdout, "=========================================================\n");
			return -1;

		default:
			fprintf(stderr, "Invalid Option. Please Choose a correct option value...\n");
			break;
		}
	}

	return 0;
}

span_data_t *sample_app_span_menu(void)
{
	int idx = 0;
	char choice[128];

	while (1) {
		printf(" Span Menu \n");
		for (idx = 1; idx <= FTDM_MAX_SPANS_INTERFACE; idx++) {

			if (!SPANS[idx].ftdm_span) {
				continue;
			}

			printf("%d. %s\n", idx, ftdm_span_get_name(SPANS[idx].ftdm_span));
		}

		printf("Select Span ID or x for exit\n");
		printf("\n");

		if (NULL != fgets(choice, 20, stdin)) {
			if (19 < strlen(choice)) {
				fprintf(stderr, "Invalid Input...\n");
				continue;
			} else {
				break;
			}
		}
	}

	choice[strlen(choice) - 1] = '\0';
	idx = atoi(choice);

	if (idx == 0 || idx > FTDM_MAX_SPANS_INTERFACE) {
		printf("Invalid ID\n");
		return sample_app_span_menu();
	}

	if (!SPANS[idx].ftdm_span) {
		return sample_app_span_menu();
	}

	printf("Got Span %s\n", ftdm_span_get_name(SPANS[idx].ftdm_span));
	printf("\n");

	return &SPANS[idx];
}

int sample_app_print_status(span_data_t * span_data)
{
	char choice[6];
	int ret = 1;

	fprintf(stdout, "\n\n\nPlease select any one of the below options :\n");

	fprintf(stdout,
			"1. Physical Span list\n2. Physical Span Status\n3. GSM Span status\n4. Go Back To Previous Menu\n");
	fprintf(stdout, "Enter your choice[1-4]: ");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			fprintf(stderr, "Invalid Input Choice must must be [1-4]...\n");
			return -1;
		}

		choice[1] = '\0';

		/* Provides information w.r.t spans configured dependinh upon the user request like:
		 * 1) List of Span configured with information about the respective span
		 * 2) Information with respect to a particular span as requested by user
		 * 3) Infotmation with respect to GSM span as requested by user
		 * 4) Go back to Previous Meanu
		 */
		switch (choice[0]) {
		case '1':
			{
				ret = sangoma_freetdm_channel_list(span_data->ftdm_span, 1);
				ftdm_sleep(100);
			}
			break;

		case '2':
			{
				ret = sangoma_freetdm_channel_dump(span_data->ftdm_span, 1);
				ftdm_sleep(100);
			}
			break;

		case '3':
			{
				ret = sangoma_freetdm_gsm_span_status(span_data->ftdm_span);
				ftdm_sleep(100);
			}
			break;

		default:
			fprintf(stderr, "Invalid Option. Please Choose a correct option value...\n");
			break;
		}
	}

	return ret;
}

static void *gsm_sample_call_echo(ftdm_thread_t * me, void *obj)
{
	uint8_t frame[1024];
	ftdm_size_t len;
	ftdm_channel_t *chan = (ftdm_channel_t *) obj;

	while (ftdm_channel_call_check_answered(chan)) {
		ftdm_wait_flag_t flags = FTDM_READ;

		if (ftdm_channel_wait(chan, &flags, -1) == FTDM_FAIL) {
			break;
		}
		len = sizeof(frame);
		if (flags & FTDM_READ) {
			if (ftdm_channel_read(chan, frame, &len) == FTDM_SUCCESS) {
				//ftdm_log(FTDM_LOG_DEBUG, "WRITE %d\n", len);
				ftdm_channel_write(chan, frame, sizeof(frame), &len);
			} else {
				break;
			}
		}
	}
	return NULL;
}

static void *app_freetdm_event_queue_handler(ftdm_thread_t * me, void *obj)
{
	ftdm_sigmsg_t *sigmsg;

	while (g_running) {
		sigmsg = ftdm_queue_dequeue(sample_app_ftdm_init.event_msg_queue);
		if (sigmsg) {
			/*
			   printf("Span = %d, chan=%d, event_id=%d, call_priv*=%p \n", 
			   sigmsg->span_id, sigmsg->chan_id, 
			   sigmsg->event_id, sigmsg->call_priv
			   );
			 */

			switch (sigmsg->event_id) {
			case FTDM_SIGEVENT_START:
				{
					ftdm_log(FTDM_LOG_INFO, "SIG_START event received. Starts io data operation for [s%dc%d]. \n",
							 sigmsg->span_id, sigmsg->chan_id);

					/* Incoming Call. Proceed to answer the call and then
					   launch a media thread that will perform I/O on the channel */
					ftdm_channel_call_answer(sigmsg->channel);
					if (ftdm_thread_create_detached(gsm_sample_call_echo, sigmsg->channel) != FTDM_SUCCESS) {
						ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
								 sigmsg->span_id, sigmsg->chan_id);
						break;
					}
					break;
				}
			case FTDM_SIGEVENT_UP:
				{
					ftdm_log(FTDM_LOG_INFO, "SIG_UP event received. Starts io data operation for [s%dc%d]. \n",
							 sigmsg->span_id, sigmsg->chan_id);

					/* Outbound call has been answered, proceed to launch a meda thread that will
					   perform I/O on the channel */
					if (ftdm_thread_create_detached(gsm_sample_call_echo, sigmsg->channel) != FTDM_SUCCESS) {
						ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
								 sigmsg->span_id, sigmsg->chan_id);
						break;
					}
					break;
				}
			case FTDM_SIGEVENT_SMS:
				{
					/* Recevied SMS message */
					ftdm_caller_data_t *caller_data = ftdm_channel_get_caller_data(sigmsg->channel);
					ftdm_sms_data_t *sms = (ftdm_sms_data_t *) caller_data->priv;
					ftdm_log(FTDM_LOG_INFO, "SMS Recieved from %s number: %s", sms->from, sms->body);
					break;
				}
			default:
				break;
			}

			/* sample app owns the event buffer. It must be deallocated here
			   otherwise memory leak */
			ftdm_safe_free(sigmsg);
		}

		/* TODO: implement wait on queue */
		ftdm_sleep(100);
	}
	return NULL;
}

int main(void)
{
	/* local variables initialization */
	int max_wait = 20;
	int err;
	int detected = 0;
	int configured = 0;
	sangoma_hw_config_reg_t reg;

	memset(&reg, 0, sizeof(reg));

#ifndef __WINDOWS__
	/* Signal handler in order to handle different signal */
	signal(SIGINT, on_sigint);
	signal(SIGUSR1, on_sigusr);
	signal(SIGUSR2, on_sigusr);
#endif

	/* Initialize global FreeTDM init structure */
	memset(&sample_app_ftdm_init, 0, sizeof(sample_app_ftdm_init));

	/* Set basic debuggin parameters */
#ifdef SDK_DEBUGGING
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_DEBUG;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_ON_ASSERT;
#else
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_INFO;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_NEVER;
#endif

	/* Initilize FreeTDM Framework 
	   Global Varialbes and FreeTDM Logger */
	err = sangoma_freetdm_init(&sample_app_ftdm_init);
	if (err) {
		return err;
	}

	/* Initialize Sangoma Config SDK */
	reg.user_obj = NULL;
	reg.hwprobe_callback = sample_app_hwprobe_callback;
	err = sangoma_config_reg(&reg);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Sangoma Config SDK Initialization Error\n");
		goto exit_sample_app;
	}

	/* Create a global FreeTDM event handling queue.
	 * FreeTDM event callback function is not allowed to exectue FreeTDM API commands.
	 * FreeTDM event callback fucntion will queue events, and the app event handler thread
	 * will consume and handle FreeTDM events.  
	 * The app event handler is allowed to run FreeTDM api commands directly */
	if (ftdm_thread_create_detached(app_freetdm_event_queue_handler, NULL) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create message queue handler thread. \n");
		goto exit_sample_app;
	}

	/* Detect Sangoma Ports 
	   For each detected port run the sample_app_hwprobe_callback
	 */
	err = sangoma_config_hwprobe(&detected, &configured, NULL);

	ftdm_log(FTDM_LOG_INFO, "Sangoma SDK Detected %d Configured %d\n", detected, configured);

	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error Failed to detect sangoma hardware \n");

		/* We cannot exit here, as the thread is launched */
		goto exit_sample_app;
	}

	if (configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Error Failed to configure any sangoma hardware \n");
		/* We cannot exit here, as the thread is launched */
		goto exit_sample_app;
	}

	/* At this point all sangoma hw ports have been configured and started.
	   And FreeTDM Spans have been creaed and started on each hw port */
	ftdm_sleep(1000);

	/* User drive main menu. 
	 * Test features for the Sangoma TDM SDK 
	 */
	while (gsm_main_menu() == 0) {
		continue;
	}

  exit_sample_app:

	/* Stop all threads. 
	 * TODO: use join instead */
	g_running = 0;
	ftdm_sleep(1000);

	while (g_num_threads && (max_wait-- > 0)) {
		fprintf(stderr, "Waiting for all threads to finish (%d)\n", g_num_threads);
		ftdm_sleep(100);
	}

	/* Stop FreeTDM Freamwork and all Spans */
	ftdm_global_destroy();

	/* TODO: Stop all Sangoma hw ports */
	ftdm_sleep(1000);
	return 0;
}
